.grid.cards {
    display: grid;
    grid-template-columns: repeat(auto-fit, minmax(10px, 1fr));
    gap: 1rem;
  }

  .grid.cards > ul {
    display: contents;
  }

  .grid.cards > ul > li {
    list-style-type: none;
  }

  .grid.cards > ul > li > a {
    display: block;
    text-decoration: none;
    color: inherit;
    border: 1px solid var(--md-default-fg-color--lightest);
    border-radius: 5px;
    padding: 1rem;
    transition: box-shadow 0.3s ease-in-out, transform 0.3s ease-in-out;
    background-color: var(--md-default-bg-color);
  }

  .grid.cards > ul > li > a:hover {
    box-shadow: 0 4px 8px var(--md-default-fg-color--lightest);
    transform: translateY(-2px);
  }

  .grid.cards > ul > li img {
    width: 100%;
    height: auto;
    margin-bottom: 0.5rem;
    object-fit: cover;
  }

  .grid.cards > ul > li strong {
    display: block;
    text-align: center;
  }
